import json
import os
import pathlib
import socketserver
from typing import Dict, Union

import gistim


class StatefulTimServer(socketserver.ThreadingTCPServer):
    """
    Stores the state of the analytic element model. If the geopackage content
    have not changed, there is no need to re-initialize the model, and solve it
    again.

    If e.g. only cellsize or domain change, values can be computed immediately
    with the headgrid function.
    """

    def __init__(self, *args, **kwargs):
        super(__class__, self).__init__(*args, **kwargs)
        self.geopackage_hash = None
        self.timml_model = None
        self.ttim_model = None
        self.solved = False


class TimHandler(socketserver.BaseRequestHandler):
    """
    The handler deals with the individual requests from the QGIS plugin.

    It will initialize the model, compute the results for a given domain
    and cellsize, and write the result to a 3D (layer, y, x) netCDF file.
    """

    def compute(
        self,
        inpath: Union[pathlib.Path, str],
        outpath: Union[pathlib.Path, str],
        mode: str,
        cellsize: float,
        active_elements: Dict[str, bool],
        as_trimesh: bool = False,
    ) -> None:
        """
        Compute the results of TimML model.

        The model is fully specified by the GeoPacakge dataset in the path.

        The extent of the head grids is read from a vector layer in the
        GeoPackage file.

        Parameters
        ----------
        path: Union[pathlib.Path, str]
            Path to the GeoPackage file containing the full model input.
        cellsize: float
            Grid cell size of the computed output

        Returns
        -------
        None
            The result is written to a netCDF file. Its name is generated from
            the geopackage name, and the requested grid cell size.
        """
        path = pathlib.Path(inpath)
        timml_spec, ttim_spec = gistim.model_specification(path, active_elements)
        self.server.timml_model, _ = gistim.timml_elements.initialize_model(timml_spec)
        self.server.timml_model.solve()

        if mode == "transient":
            self.server.ttim_model, _ = gistim.ttim_elements.initialize_model(
                ttim_spec, self.server.timml_model
            )

        extent, crs = gistim.gridspec(path, cellsize)

        if mode == "steady-state":
            if as_trimesh:
                ugrid_head = gistim.timml_elements.headmesh(
                    self.server.timml_model, timml_spec, cellsize
                )
            else:
                head = gistim.timml_elements.headgrid(
                    self.server.timml_model, extent, cellsize
                )
                ugrid_head = gistim.to_ugrid2d(head)

        elif mode == "transient":
            print("Solving transient model")
            self.server.ttim_model.solve()
            if as_trimesh:
                ugrid_head = gistim.ttim_elements.headmesh(
                    self.server.ttim_model, timml_spec, cellsize
                )
            else:
                head = gistim.ttim_elements.headgrid(
                    self.server.ttim_model,
                    extent,
                    cellsize,
                    ttim_spec.output_times,
                    ttim_spec.temporal_settings["reference_date"].iloc[0],
                )
                ugrid_head = gistim.to_ugrid2d(head)

        else:
            raise ValueError(
                f'Mode should be "steady-state" or "transient". Received: {mode}'
            )

        print("Writing result to:", outpath)
        ugrid_head.to_netcdf(outpath)

    def handle(self) -> None:
        """
        Handle a request. This function has to be overloaded for a request
        handler class.
        """
        # TODO: rfile stream? Seems more robust than these 1024 bytes
        # TODO: try-except, and return error in return message
        message = self.request.recv(1024 * 1024).strip()
        data = json.loads(message)

        print("JSON received:")
        print(json.dumps(data, indent=4))

        operation = data.pop("operation")

        if operation == "compute":
            self.compute(
                inpath=data["inpath"],
                outpath=data["outpath"],
                cellsize=data["cellsize"],
                mode=data["mode"],
                active_elements=data["active_elements"],
                as_trimesh=data["as_trimesh"],
            )
            print("Computation succesful\n\n")
            # Send error code 0: all okay
            self.request.sendall(bytes("0", "utf-8"))

        elif operation == "process_ID":
            self.request.sendall(bytes(str(os.getpid()), "utf-8"))

        elif operation == "convert":
            inpath = data["inpath"]
            outpath = data["outpath"]
            gistim.convert_to_script(inpath, outpath)
            self.request.sendall(bytes("0", "utf-8"))

        elif operation == "extract":
            inpath = data["inpath"]
            outpath = data["outpath"]
            wkt_geometry = data["wkt_geometry"].split(";")
            gistim.data_extraction.netcdf_to_table(
                inpath=inpath,
                outpath=outpath,
                wkt_geometry=wkt_geometry,
            )
            print(f"Extraction from {inpath} to {outpath} succesful\n\n")
            # Send error code 0: all okay
            self.request.sendall(bytes("0", "utf-8"))

        else:
            print('Invalid operation. Valid options are: "compute", "process_ID".')



"""
This model converts the geodataframe as read from the geopackage into keyword
arguments for TTim.

These keyword arguments are used to initialize a model, or used to generate a
Python script.

"""
from typing import Any, Dict, List, Tuple

import black
import geopandas as gpd
import numpy as np
import pandas as pd
import timml
import tqdm
import xarray as xr

try:
    import ttim
    from ttim.model import TimModel
except ImportError:
    ttim = None
    TimModel = None

from . import ugrid
from .common import (
    ElementSpecification,
    FloatArray,
    TransientElementSpecification,
    TtimModelSpecification,
    aquifer_data,
    dict_to_kwargs_code,
    headgrid_code,
    linestring_coordinates,
    point_coordinates,
    sanitized,
    trimesh,
)


# Dataframe to ttim element
# --------------------------
def transient_dataframe(spec):
    if spec.dataframe is not None:
        return spec.dataframe.set_index("geometry_id")
    else:
        return None


def transient_input(timeseries_df, row, field, tstart) -> List:
    geometry_id = row["geometry_id"]
    if timeseries_df is None or geometry_id not in timeseries_df.index:
        return [(tstart, 0.0)]
    else:
        timeseries = timeseries_df.loc[[geometry_id]]
        timeseries[field] -= row[field]
        timeseries = timeseries.sort_values(by="tstart")
        return list(timeseries[["tstart", field]].itertuples(index=False, name=None))


def ttim_model(
    dataframe: gpd.GeoDataFrame,
    temporal_settings: pd.DataFrame,
    timml_model: timml.Model,
) -> TimModel:
    """
    Parameters
    ----------
    dataframe: geopandas.GeoDataFrame

    Returns
    -------
    ttim.TimModel
    """
    data = aquifer_data(dataframe, transient=True)
    for arg in ("tstart", "tmin", "tmax", "M"):
        data[arg] = temporal_settings[arg].iloc[0]
    data["timmlmodel"] = timml_model
    return data


def well(spec: TransientElementSpecification, tstart: float) -> List[Dict[str, Any]]:
    df = transient_dataframe(spec)
    dataframe = spec.steady_spec.dataframe
    X, Y = point_coordinates(dataframe)
    kwargslist = []
    for (row, x, y) in zip(dataframe.to_dict("records"), X, Y):
        kwargslist.append(
            {
                "xw": x,
                "yw": y,
                "tsandQ": transient_input(df, row, "discharge", tstart),
                "rw": row["radius"],
                "res": row["resistance"],
                "layers": row["layer"],
                "label": row["label"],
                "rc": row["caisson_radius"],
                "wbstype": "slug" if row["slug"] else "pumping",
            }
        )
    return kwargslist


def headwell(
    spec: TransientElementSpecification, tstart: float
) -> List[Dict[str, Any]]:
    df = transient_dataframe(spec)
    dataframe = spec.steady_spec.dataframe
    X, Y = point_coordinates(dataframe)
    kwargslist = []
    for (row, x, y) in zip(dataframe.to_dict("records"), X, Y):
        kwargslist.append(
            {
                "xw": x,
                "yw": y,
                "tsandh": transient_input(df, row, "head", tstart),
                "rw": row["radius"],
                "res": row["resistance"],
                "layers": row["layer"],
                "label": row["label"],
            }
        )
    return kwargslist


def headlinesink(
    spec: TransientElementSpecification, tstart: float
) -> List[Dict[str, Any]]:
    df = transient_dataframe(spec)
    kwargslist = []
    for row in spec.steady_spec.dataframe.to_dict("records"):
        kwargslist.append(
            {
                "xy": linestring_coordinates(row),
                "tsandh": transient_input(df, row, "head", tstart),
                "res": row["resistance"],
                "wh": row["width"],
                "layers": row["layer"],
                "label": row["label"],
            }
        )
    return kwargslist


def linesinkditch(spec: ElementSpecification, tstart: float) -> List[Dict[str, Any]]:
    df = transient_dataframe(spec)
    kwargslist = []
    for row in spec.steady_spec.dataframe.to_dict("records"):
        kwargslist.append(
            {
                "xy": linestring_coordinates(row),
                "tsandQ": transient_input(df, row, "discharge", tstart),
                "res": row["resistance"],
                "wh": row["width"],
                "layers": row["layer"],
                "label": row["label"],
            }
        )
    return kwargslist


def leakylinedoublet(spec: ElementSpecification, tstart: float) -> List[Dict[str, Any]]:
    kwargslist = []
    for row in spec.dataframe.to_dict("records"):
        kwargslist.append(
            {
                "xy": linestring_coordinates(row),
                "res": row["resistance"],
                "layers": row["layer"],
                "order": row["order"],
                "label": row["label"],
            }
        )
    return kwargslist


def implinedoublet(spec: ElementSpecification, tstart: float) -> List[Dict[str, Any]]:
    kwargslist = []
    for row in spec.dataframe.to_dict("records"):
        kwargslist.append(
            {
                "xy": linestring_coordinates(row),
                "res": "imp",
                "layers": row["layer"],
                "order": row["order"],
                "label": row["label"],
            }
        )
    return kwargslist


def circareasink(spec: TransientElementSpecification, tstart: float) -> None:
    """
    Parameters
    ----------
    dataframe: geopandas.GeoDataFrame

    Returns
    -------
    None
    """
    df = transient_dataframe(spec)
    kwargslist = []
    for i, row in spec.steady_spec.dataframe.to_dict("records"):
        x, y = row.geometry.centroid.xy
        coords = np.array(row.geometry.exterior.coords)
        x0, y0 = coords[0]
        radius = np.sqrt((x0 - x) ** 2 + (y0 - y) ** 2)
        kwargslist.append(
            {
                "xc": x,
                "yc": y,
                "R": radius,
                "tsandN": transient_input(df, row, "rate", tstart),
            }
        )
    return kwargslist


def headgrid(
    model: TimModel,
    extent: Tuple[float],
    cellsize: float,
    times: FloatArray,
    reference_date: pd.Timestamp,
) -> xr.DataArray:
    xmin, xmax, ymin, ymax = extent
    x = np.arange(xmin, xmax, cellsize) + 0.5 * cellsize
    # In geospatial rasters, y is DECREASING with row number
    y = np.arange(ymax, ymin, -cellsize) - 0.5 * cellsize
    nlayer = model.aq.find_aquifer_data(x[0], y[0]).naq
    layer = [i for i in range(nlayer)]
    head = np.empty((nlayer, times.size, y.size, x.size), dtype=np.float64)
    for i in tqdm.tqdm(range(y.size)):
        for j in range(x.size):
            head[:, :, i, j] = model.head(x[j], y[i], times, layer)

    time = pd.to_datetime(reference_date) + pd.to_timedelta(times, "D")
    return xr.DataArray(
        data=head,
        name="head",
        coords={"layer": layer, "time": time, "y": y, "x": x},
        dims=("layer", "time", "y", "x"),
    )


def headmesh(
    model: TimModel,
    spec,
    cellsize: float,
    times: FloatArray,
    reference_date: pd.Timestamp,
) -> xr.Dataset:
    nodes, face_nodes, centroids = trimesh(spec, cellsize)
    nface = len(face_nodes)

    nlayer = model.aq.find_aquifer_data(nodes[0, 0], nodes[0, 0]).naq
    layer = [i for i in range(nlayer)]
    head = np.empty((nlayer, times.size, nface), dtype=np.float64)
    for i in tqdm.tqdm(range(nface)):
        x = centroids[i, 0]
        y = centroids[i, 1]
        head[:, :, i] = model.head(x, y, times, layer)
    uds = ugrid._ugrid2d_dataset(
        node_x=nodes[:, 0],
        node_y=nodes[:, 1],
        face_x=centroids[:, 0],
        face_y=centroids[:, 1],
        face_nodes=face_nodes,
    )
    time = pd.to_datetime(reference_date) + pd.to_timedelta(times, "D")
    uds = uds.assign_coords(time=time)
    uds["head"] = xr.DataArray(head, dims=("layer", "time", "face"))
    return ugrid._unstack_layers(uds)


# Map the names of the elements to their constructors
if ttim is not None:
    MAPPING = {
        "Circular Area Sink": (circareasink, ttim.CircAreaSink),
        "Well": (well, ttim.Well),
        "Head Well": (headwell, ttim.HeadWell),
        "Head Line Sink": (headlinesink, ttim.HeadLineSinkString),
        "Line Sink Ditch": (linesinkditch, ttim.LineSinkDitchString),
        "Leaky Line Doublet": (leakylinedoublet, ttim.LeakyLineDoubletString),
        "Impermeable Line Doublet": (implinedoublet, ttim.LeakyLineDoubletString),
    }


def initialize_model(spec: TtimModelSpecification, timml_model) -> TimModel:
    """
    Initialize a Ttim analytic model based on the data in a geopackage.

    Parameters
    ----------
    spec: ModelSpecification
        Named tuple with the layer name of the aquifer and a dictionary of
        other element names to its construction function.

    Returns
    -------
    timml.Model

    Examples
    --------

    >>> import gistim
    >>> path = "my-model.gpkg"
    >>> spec = gistim.model_specification(path)
    >>> model = gistim.initialize_model(spec)
    >>> model.solve()

    """
    model = ttim.ModelMaq(
        **ttim_model(spec.aquifer, spec.temporal_settings, timml_model)
    )
    elements = {}
    for name, element_spec in spec.elements.items():
        elementtype = element_spec.elementtype
        if not element_spec.active or elementtype not in MAPPING:
            continue

        print(f"adding {name} as {elementtype}")
        f_to_kwargs, element = MAPPING[elementtype]
        for i, kwargs in enumerate(f_to_kwargs(element_spec, model.tstart)):
            kwargs["model"] = model
            elements[f"{name}_{i}"] = element(**kwargs)

    return model, elements


def convert_to_script(spec: TtimModelSpecification) -> str:
    """
    Convert model specification to an equivalent Python script.
    """
    modelkwargs = ttim_model(spec.aquifer, spec.temporal_settings, "model")
    strkwargs = dict_to_kwargs_code(modelkwargs)
    strings = ["import ttim", f"ttim_model = ttim.ModelMaq({strkwargs})"]
    for name, element_spec in spec.elements.items():
        elementtype = element_spec.elementtype
        if elementtype not in MAPPING:
            continue
        print(f"adding {name} as {elementtype}")

        f_to_kwargs, element = MAPPING[elementtype]
        for i, kwargs in enumerate(f_to_kwargs(element_spec, modelkwargs["tstart"])):
            kwargs["model"] = "ttim_model"
            kwargs = dict_to_kwargs_code(kwargs)
            strings.append(
                f"ttim_{sanitized(name)}_{i} = ttim.{element.__name__}({kwargs})"
            )

    strings.append("ttim_model.solve()")

    xg, yg = headgrid_code(spec.domain)
    times = spec.output_times.tolist()
    strings.append(f"head = ttim_model.headgrid(xg={xg}, yg={yg}, t={times})")

    return black.format_str("\n".join(strings), mode=black.FileMode())



"""
Common utilities used by conversions from GeoPackage (GPKG) layers to TimML and
TTim elements.

The relation between a GPKG layer and the elements is not perfect. A GPGK layer
consists of a table, with optionally an associated geometry for every row. This
matches one to one for elements such as HeadLineSinkStrings, Wells, etc: for
these elements, one row equals one element.

For elements such as PolygonImhomogenities, this is not the case. Every geometry
(a polygon) requires a table of its own. These tables are stored in associated
tables; their association is by name.

For transient (TTim) elements, the same is true: elements require an additional
table for their timeseries data, which should require repeating the geometry
for every time step. In this package, we assume that any ttim element is
accompanied by a timml element; the QGIS plugin always sets up the layer ih
that manner.

When processing a GPKG, we first parse the names, and group the different
tables together in the ElementSpecifications below. The ``timml_elements`` and
``ttim_elements`` then convert these grouped tables into ``timml`` and ``ttim``
models.
"""
import pathlib
import re
from collections import defaultdict
from functools import partial
from typing import Any, Dict, NamedTuple, Tuple, Union

import fiona
import geopandas as gpd
import numpy as np
import pandas as pd

gpd.options.use_pygeos = False
FloatArray = np.ndarray


class ElementSpecification(NamedTuple):
    elementtype: str
    active: bool
    dataframe: gpd.GeoDataFrame
    associated_dataframe: gpd.GeoDataFrame


class TransientElementSpecification(NamedTuple):
    elementtype: str
    active: bool
    dataframe: gpd.GeoDataFrame
    steady_spec: ElementSpecification


class TimmlModelSpecification(NamedTuple):
    aquifer: gpd.GeoDataFrame
    elements: Dict[str, ElementSpecification]
    domain: gpd.GeoDataFrame


class TtimModelSpecification(NamedTuple):
    aquifer: gpd.GeoDataFrame
    temporal_settings: gpd.GeoDataFrame
    elements: Dict[str, ElementSpecification]
    domain: gpd.GeoDataFrame
    output_times: FloatArray


# Extract coordinates from geodataframe
# -------------------------------------
def point_coordinates(dataframe) -> Tuple[FloatArray, FloatArray]:
    return dataframe["geometry"].x, dataframe["geometry"].y


def linestring_coordinates(row) -> Tuple[FloatArray, FloatArray]:
    return np.array(row["geometry"].coords)


def polygon_coordinates(row) -> Tuple[FloatArray, FloatArray]:
    return np.array(row["geometry"].exterior.coords)


# Parse GPKG content to Tim input
# -------------------------------
def aquifer_data(
    dataframe: gpd.GeoDataFrame, transient: bool = False
) -> Dict[str, Any]:
    """
    Convert a table created by the QGIS plugin to the layer configuration and
    keywords arguments as expected by TimML or TTim.
    """
    # Make sure the layers are in the right order.
    dataframe = dataframe.sort_values(by="layer").set_index("layer")
    nlayer = len(dataframe)
    # Deal with optional semi-confined top layer.
    hstar = dataframe.loc[0, "semiconf_head"]
    semi = pd.notnull(hstar)
    kaq = dataframe["aquifer_k"].values

    if semi:
        c = dataframe["aquitard_c"].values
        porosity = np.empty(nlayer * 2)
        z = np.empty(nlayer * 2 + 1)
        z[0] = dataframe.loc[0, "semiconf_top"]
        z[1::2] = dataframe["aquifer_top"].values
        z[2::2] = dataframe["aquifer_bottom"].values
        porosity[::2] = dataframe["aquitard_npor"].values
        porosity[1::2] = dataframe["aquifer_npor"].values
        topboundary = "semi"
        storage_aquifer = dataframe["aquifer_s"].values
        storage_aquitard = dataframe["aquitard_s"].values
    else:
        c = dataframe["aquitard_c"].values[1:]
        z = np.empty(nlayer * 2)
        z[::2] = dataframe["aquifer_top"].values
        z[1::2] = dataframe["aquifer_bottom"].values
        porosity = np.empty(nlayer * 2 - 1)
        porosity[::2] = dataframe["aquifer_npor"].values
        porosity[1::2] = dataframe["aquitard_npor"].values[1:]
        topboundary = "conf"
        storage_aquifer = dataframe["aquifer_s"].values
        storage_aquitard = dataframe["aquifer_s"].values[1:]

    d = {
        "kaq": kaq,
        "z": z,
        "c": c,
        "topboundary": topboundary,
    }
    if transient:
        d["Sll"] = storage_aquitard
        d["Saq"] = storage_aquifer
        # TODO: for now, assume there is always at least one specific yield
        # this is the aquifer if conf, the aquitard if semi
        d["phreatictop"] = True
    else:
        d["npor"] = porosity
        d["hstar"] = hstar

    return d


def parse_name(layername: str) -> Tuple[str, str, str]:
    """
    Based on the layer name find out:

    * whether it's a timml or ttim element;
    * which element type it is;
    * what the user provided name is.

    For example:
    parse_name("timml Headwell: drainage") -> ("timml", "Head Well", "drainage")

    Some grouping of tables occurs here.
    """
    prefix, name = layername.split(":")
    element_type = re.split("timml |ttim ", prefix)[1]
    mapping = {
        "Computation Times": "Domain",
        "Temporal Settings": "Aquifer",
        "Polygon Inhomogeneity Properties": "Polygon Inhomogeneity",
        "Building Pit Properties": "Building Pit",
    }
    element_type = mapping.get(element_type, element_type)
    if "timml" in prefix:
        if "Properties" in prefix:
            tim_type = "timml_assoc"
        else:
            tim_type = "timml"
    elif "ttim" in prefix:
        tim_type = "ttim"
    else:
        raise ValueError(f"Neither timml nor ttim in layername: {layername}")
    return tim_type, element_type, name


def model_specification(
    path: Union[str, pathlib.Path], active_elements: Dict[str, bool]
) -> Tuple[TimmlModelSpecification, TtimModelSpecification]:
    """
    Group the different layers of a GPKG into model specifications for timml
    and ttim. The grouping occurs solely on the basis of layer names.
    """
    # Start by listing all layers
    gpkg_names = fiona.listlayers(path)
    # Group all these names together (using a defaultdict)
    dd = defaultdict
    grouped_names = dd(partial(dd, partial(dd, list)))
    for layername in gpkg_names:
        tim_type, element_type, name = parse_name(layername)
        grouped_names[element_type][name][tim_type] = layername

    # Grab the names of the required elements and load the data into
    # geodataframes.
    aquifer_entry = grouped_names.pop("Aquifer")["Aquifer"]
    aquifer = gpd.read_file(path, layer=aquifer_entry["timml"])
    temporal_settings = gpd.read_file(path, layer=aquifer_entry["ttim"])
    domain_entry = grouped_names.pop("Domain")["Domain"]
    domain = gpd.read_file(path, layer=domain_entry["timml"])
    output_times = gpd.read_file(path, layer=domain_entry["ttim"])

    # Load the data all other elements into geodataframes.
    ttim_elements = {}
    timml_elements = {}
    for element_type, element_group in grouped_names.items():
        for name, group in element_group.items():
            timml_name = group["timml"]
            timml_assoc_name = group.get("timml_assoc", None)
            ttim_name = group.get("ttim", timml_name)

            timml_df = gpd.read_file(path, layer=timml_name)
            timml_assoc_df = (
                gpd.read_file(path, layer=timml_assoc_name)
                if timml_assoc_name is not None
                else None
            )
            ttim_df = (
                gpd.read_file(path, layer=ttim_name) if ttim_name is not None else None
            )
            timml_spec = ElementSpecification(
                elementtype=element_type,
                active=active_elements.get(timml_name, False),
                dataframe=timml_df,
                associated_dataframe=timml_assoc_df,
            )

            ttim_spec = TransientElementSpecification(
                elementtype=element_type,
                active=active_elements.get(ttim_name, False),
                dataframe=ttim_df,
                steady_spec=timml_spec,
            )
            timml_elements[timml_name] = timml_spec
            ttim_elements[ttim_name] = ttim_spec

    return (
        TimmlModelSpecification(aquifer, timml_elements, domain),
        TtimModelSpecification(
            aquifer,
            temporal_settings,
            ttim_elements,
            domain,
            np.sort(output_times["time"].values),
        ),
    )


# Three helpers for conversion to Python scripts
# ----------------------------------------------
def dict_to_kwargs_code(data: dict) -> str:
    strings = []
    for key, value in data.items():
        if isinstance(value, np.ndarray):
            value = value.tolist()
        elif isinstance(value, str) and key not in ("model", "timmlmodel"):
            value = f'"{value}"'
        strings.append(f"{key}={value}")
    return ",".join(strings)


def sanitized(name: str) -> str:
    return name.split(":")[-1].replace(" ", "_")


def headgrid_code(domain: gpd.GeoDataFrame) -> Tuple[str, str]:
    xmin, ymin, xmax, ymax = domain.bounds.iloc[0]
    dy = (ymax - ymin) / 50.0
    if dy > 500.0:
        dy = round(dy / 500.0) * 500.0
    elif dy > 50.0:
        dy = round(dy / 50.0) * 50.0
    elif dy > 5.0:  # round to five
        dy = round(dy / 5.0) * 5.0
    elif dy > 1.0:
        dy = round(dy)
    (xmin, xmax, ymin, ymax) = round_extent((xmin, xmax, ymin, ymax), dy)
    xmin += 0.5 * dy
    xmax += 0.5 * dy
    ymax -= 0.5 * dy
    xmin -= 0.5 * dy
    xg = f"np.arange({xmin}, {xmax}, {dy})"
    yg = f"np.arange({ymax}, {ymin}, -{dy})"
    return xg, yg


# Output methods
# --------------
def round_extent(extent: Tuple[float], cellsize: float) -> Tuple[float]:
    """
    Increases the extent until all sides lie on a coordinate
    divisible by cellsize.

    Parameters
    ----------
    extent: Tuple[float]
        xmin, xmax, ymin, ymax
    cellsize: float
        Desired cell size of the output head grids

    Returns
    -------
    extent: Tuple[float]
        xmin, xmax, ymin, ymax
    """
    xmin, xmax, ymin, ymax = extent
    xmin = np.floor(xmin / cellsize) * cellsize
    ymin = np.floor(ymin / cellsize) * cellsize
    xmax = np.ceil(xmax / cellsize) * cellsize
    ymax = np.ceil(ymax / cellsize) * cellsize
    return xmin, xmax, ymin, ymax


def gridspec(
    path: Union[pathlib.Path, str], cellsize: float
) -> Tuple[Tuple[float], Any]:
    """
    Infer the grid specification from the geopackage ``timmlDomain`` layer and
    the provided cellsize.

    Parameters
    ----------
    path: Union[pathlib.Path, str]
        Path to the GeoPackage file.
    cellsize: float
        Desired cell size of the output head grids

    Returns
    -------
    extent: tuple[float]
        xmin, xmax, ymin, ymax
    crs: Any
        Coordinate Reference System
    """
    domain = gpd.read_file(path, layer="timml Domain:Domain")
    xmin, ymin, xmax, ymax = domain.bounds.iloc[0]
    extent = (xmin, xmax, ymin, ymax)
    return round_extent(extent, cellsize), domain.crs


def trimesh(spec: TimmlModelSpecification, cellsize: float):
    # Only import it if needed
    import geomesh

    domain = spec.domain

    geometry = []
    for spec in spec.elements.values():
        df = spec.dataframe
        if spec.elementtype in ["Well", "HeadWell"]:
            df.geometry = df.buffer(cellsize, resolution=2)
        if spec.elementtype in ["PolygonInhomogeneity", "Building Pit"]:
            pass
        elif spec.elementtype in ["Impermeable Line Doublet", "Leaky Line Doublet"]:
            df.geometry = df.buffer(cellsize * 0.01, cap_style=2, join_style=2)
        geometry.append(
            gpd.overlay(df, domain, how="intersection").loc[:, ["geometry"]]
        )

    gdf = pd.concat([domain.loc[:, ["geometry"]], *geometry])
    gdf["cellsize"] = cellsize
    # intersect bounding box with elements
    mesher = geomesh.TriangleMesher(gdf)
    mesher.minimum_angle = 0.01
    mesher.conforming_delaunay = False
    nodes, face_nodes = mesher.generate()
    centroids = nodes[face_nodes].mean(axis=1)
    return nodes, face_nodes, centroids



"""
This model converts the geodataframe as read from the geopackage into keyword
arguments for TimML.

These keyword arguments are used to initialize a model, or used to generate a
Python script.
"""
from typing import Any, Dict, List, Tuple

import black
import numpy as np
import timml
import tqdm
import xarray as xr

from . import ugrid
from .common import (
    ElementSpecification,
    TimmlModelSpecification,
    aquifer_data,
    dict_to_kwargs_code,
    headgrid_code,
    linestring_coordinates,
    point_coordinates,
    polygon_coordinates,
    sanitized,
    trimesh,
)


def constant(spec: ElementSpecification) -> List[Dict[str, Any]]:
    firstrow = spec.dataframe.iloc[0]
    x, y = point_coordinates(firstrow)
    return [
        {
            "xr": x,
            "yr": y,
            "hr": firstrow["head"],
            "layer": firstrow["layer"],
            "label": firstrow["label"],
        }
    ]


def uflow(spec: ElementSpecification) -> List[Dict[str, Any]]:
    row = spec.dataframe.iloc[0]
    return [
        {
            "slope": row["slope"],
            "angle": row["angle"],
            "label": row["label"],
        }
    ]


def well(spec: ElementSpecification) -> List[Dict[str, Any]]:
    dataframe = spec.dataframe
    X, Y = point_coordinates(dataframe)
    kwargslist = []
    for (row, x, y) in zip(dataframe.to_dict("records"), X, Y):
        kwargslist.append(
            {
                "xw": x,
                "yw": y,
                "Qw": row["discharge"],
                "rw": row["radius"],
                "res": row["resistance"],
                "layers": row["layer"],
                "label": row["label"],
            }
        )
    return kwargslist


def headwell(spec: ElementSpecification) -> List[Dict[str, Any]]:
    dataframe = spec.dataframe
    X, Y = point_coordinates(dataframe)
    kwargslist = []
    for (row, x, y) in zip(dataframe.to_dict("records"), X, Y):
        kwargslist.append(
            {
                "xw": x,
                "yw": y,
                "hw": row["head"],
                "rw": row["radius"],
                "res": row["resistance"],
                "layers": row["layer"],
                "label": row["label"],
            }
        )
    return kwargslist


def polygoninhom(spec: ElementSpecification) -> List[Dict[str, Any]]:
    """
    Parameters
    ----------
    dataframe: tuple of geopandas.GeoDataFrame

    Returns
    -------
    None
    """
    geometry = spec.dataframe
    properties = spec.associated_dataframe.set_index("geometry_id")
    # Iterate through the row containing the geometry
    # and iterate through the associated table containing k properties.
    kwarglist = []
    for row in geometry.to_dict("records"):
        dataframe = properties.loc[[row["geometry_id"]]]
        kwargs = aquifer_data(dataframe)
        kwargs["xy"] = polygon_coordinates(row)
        kwargs["order"] = row["order"]
        kwargs["ndeg"] = row["ndegrees"]
        kwarglist.append(kwargs)
    return kwarglist


def buildingpit(spec: ElementSpecification) -> List[Dict[str, Any]]:
    geometry = spec.dataframe
    properties = spec.associated_dataframe.set_index("geometry_id")
    # Iterate through the row containing the geometry
    # and iterate through the associated table containing k properties.
    kwarglist = []
    for row in geometry.to_dict("records"):
        dataframe = properties.loc[row["geometry_id"]]
        kwargs = aquifer_data(dataframe)
        kwargs["xy"] = polygon_coordinates(row)
        kwargs["order"] = row["order"]
        kwargs["ndeg"] = row["ndegrees"]
        kwargs["layers"] = np.atleast_1d(row["layer"])
        kwarglist.append(kwargs)
    return kwarglist


def headlinesink(spec: ElementSpecification) -> List[Dict[str, Any]]:
    kwargslist = []
    for row in spec.dataframe.to_dict("records"):
        kwargslist.append(
            {
                "xy": linestring_coordinates(row),
                "hls": row["head"],
                "res": row["resistance"],
                "wh": row["width"],
                "order": row["order"],
                "layers": row["layer"],
                "label": row["label"],
            }
        )
    return kwargslist


def linesinkditch(spec: ElementSpecification) -> List[Dict[str, Any]]:
    kwargslist = []
    for row in spec.dataframe.to_dict("records"):
        kwargslist.append(
            {
                "xy": linestring_coordinates(row),
                "Qls": row["discharge"],
                "res": row["resistance"],
                "wh": row["width"],
                "order": row["order"],
                "layers": row["layer"],
                "label": row["label"],
            }
        )
    return kwargslist


def leakylinedoublet(spec: ElementSpecification) -> List[Dict[str, Any]]:
    kwargslist = []
    for row in spec.dataframe.to_dict("records"):
        kwargslist.append(
            {
                "xy": linestring_coordinates(row),
                "res": row["resistance"],
                "layers": row["layer"],
                "order": row["order"],
                "label": row["label"],
            }
        )
    return kwargslist


def implinedoublet(spec: ElementSpecification) -> List[Dict[str, Any]]:
    kwargslist = []
    for row in spec.dataframe.to_dict("records"):
        kwargslist.append(
            {
                "xy": linestring_coordinates(row),
                "layers": row["layer"],
                "order": row["order"],
                "label": row["label"],
            }
        )
    return kwargslist


def circareasink(spec: ElementSpecification) -> List[Dict[str, Any]]:
    kwargslist = []
    for row in spec.dataframe.to_dict("records"):
        x, y = row.geometry.centroid.xy
        coords = np.array(row.geometry.exterior.coords)
        x0, y0 = coords[0]
        radius = np.sqrt((x0 - x) ** 2 + (y0 - y) ** 2)
        kwargslist.append(
            {
                "xc": x,
                "yc": y,
                "R": radius,
                "N": row["rate"],
            }
        )
    return kwargslist


def validate(spec: TimmlModelSpecification) -> None:
    if spec.aquifer is None:
        raise ValueError("Aquifer entry is missing")
    # TODO: more checks


def headgrid(model: timml.Model, extent: Tuple[float], cellsize: float) -> xr.DataArray:
    """
    Compute the headgrid of the TimML model, and store the results
    in an xarray DataArray with the appropriate dimensions.

    Parameters
    ----------
    model: timml.Model
        Solved model to get heads from
    extent: Tuple[float]
        xmin, xmax, ymin, ymax
    cellsize: float
        Desired cell size of the output head grids

    Returns
    -------
    head: xr.DataArray
        DataArray with dimensions ``("layer", "y", "x")``.
    """
    xmin, xmax, ymin, ymax = extent
    x = np.arange(xmin, xmax, cellsize) + 0.5 * cellsize
    # In geospatial rasters, y is DECREASING with row number
    y = np.arange(ymax, ymin, -cellsize) - 0.5 * cellsize
    nlayer = model.aq.find_aquifer_data(x[0], y[0]).naq
    layer = [i for i in range(nlayer)]
    head = np.empty((nlayer, y.size, x.size), dtype=np.float64)
    for i in tqdm.tqdm(range(y.size)):
        for j in range(x.size):
            head[:, i, j] = model.head(x[j], y[i], layer)

    return xr.DataArray(
        data=head,
        name="head",
        coords={"layer": layer, "y": y, "x": x},
        dims=("layer", "y", "x"),
    )


def headmesh(
    model: timml.Model, spec: TimmlModelSpecification, cellsize: float
) -> xr.Dataset:
    nodes, face_nodes, centroids = trimesh(spec, cellsize)
    nlayer = model.aq.find_aquifer_data(nodes[0, 0], nodes[0, 0]).naq
    layer = [i for i in range(nlayer)]
    head = np.empty((nlayer, len(nodes)), dtype=np.float64)
    # for i in tqdm.tqdm(range(nface)):
    #    x = centroids[i, 0]
    #    y = centroids[i, 1]
    for i, (x, y) in enumerate(tqdm.tqdm(nodes)):
        head[:, i] = model.head(x, y, layer)
        head[:, i] = model.head(x, y, layer)
    uds = ugrid._ugrid2d_dataset(
        node_x=nodes[:, 0],
        node_y=nodes[:, 1],
        face_nodes=face_nodes,
        face_x=centroids[:, 0],
        face_y=centroids[:, 1],
    )
    uds["head"] = xr.DataArray(head, dims=("layer", "node"))
    return ugrid._unstack_layers(uds)


# Map the names of the elements to their constructors
MAPPING = {
    "Constant": (constant, timml.Constant),
    "Uniform Flow": (uflow, timml.Uflow),
    "Circular Area Sink": (circareasink, timml.CircAreaSink),
    "Well": (well, timml.Well),
    "Head Well": (headwell, timml.HeadWell),
    "Polygon Inhomogeneity": (polygoninhom, timml.PolygonInhomMaq),
    "Head Line Sink": (headlinesink, timml.HeadLineSinkString),
    "Line Sink Ditch": (linesinkditch, timml.LineSinkDitchString),
    "Leaky Line Doublet": (leakylinedoublet, timml.LeakyLineDoubletString),
    "Impermeable Line Doublet": (implinedoublet, timml.ImpLineDoubletString),
    "Building Pit": (buildingpit, timml.BuildingPit),
}


def initialize_model(spec: TimmlModelSpecification) -> timml.Model:
    """
    Initialize a TimML analytic model based on the data in a geopackage.

    Parameters
    ----------
    spec: ModelSpecification
        Named tuple with the layer name of the aquifer and a dictionary of
        other element names to its construction function.

    Returns
    -------
    timml.Model

    Examples
    --------

    >>> import gistim
    >>> path = "my-model.gpkg"
    >>> spec = gistim.model_specification(path)
    >>> model = gistim.initialize_model(spec)
    >>> model.solve()

    """
    validate(spec)
    model = timml.ModelMaq(**aquifer_data(spec.aquifer))
    elements = {}
    for name, element_spec in spec.elements.items():
        if not element_spec.active:
            continue

        elementtype = element_spec.elementtype
        print(f"adding {name} as {elementtype}")

        try:
            f_to_kwargs, element = MAPPING[elementtype]
            for i, kwargs in enumerate(f_to_kwargs(element_spec)):
                kwargs["model"] = model
                elements[f"{name}_{i}"] = element(**kwargs)
        except KeyError as e:
            msg = (
                f'Invalid element specification "{elementtype}". '
                f'Available types are: {", ".join(MAPPING.keys())}.'
            )
            raise KeyError(msg) from e

    return model, elements


def convert_to_script(spec: TimmlModelSpecification) -> str:
    """
    Convert model specification to an equivalent Python script.
    """
    modelkwargs = dict_to_kwargs_code(aquifer_data(spec.aquifer))
    strings = [
        "from numpy import nan",
        "import numpy as np",
        "import timml",
        f"model = timml.ModelMaq({modelkwargs})",
    ]
    for name, element_spec in spec.elements.items():
        elementtype = element_spec.elementtype
        print(f"adding {name} as {elementtype}")

        try:
            f_to_kwargs, element = MAPPING[elementtype]
            for i, kwargs in enumerate(f_to_kwargs(element_spec)):
                kwargs["model"] = "model"
                kwargs = dict_to_kwargs_code(kwargs)
                strings.append(
                    f"{sanitized(name)}_{i} = timml.{element.__name__}({kwargs})"
                )
        except KeyError as e:
            msg = (
                f'Invalid element specification "{elementtype}". '
                f'Available types are: {", ".join(MAPPING.keys())}.'
            )
            raise KeyError(msg) from e

    strings.append("model.solve()")

    xg, yg = headgrid_code(spec.domain)
    strings.append(f"head = model.headgrid(xg={xg}, yg={yg})")

    return black.format_str("\n".join(strings), mode=black.FileMode())



import argparse
import json
import os
import platform
import sys
from pathlib import Path

import gistim


def configure(args) -> None:
    """
    Write all the environmental variables so the QGIS interpreter
    can (re)set them properly.
    """
    if platform.system() == "Windows":
        configdir = Path(os.environ["APPDATA"]) / "qgis-tim"
    else:
        configdir = Path(os.environ["HOME"]) / ".qgis-tim"
    configdir.mkdir(exist_ok=True)

    env = {key: value for key, value in os.environ.items()}
    path = configdir / "environmental-variables.json"
    if path.exists():
        with open(path, "r") as f:
            content = json.loads(f.read())
        content[sys.executable] = env
    else:
        content = {sys.executable: env}

    with open(configdir / "environmental-variables.json", "w") as f:
        f.write(json.dumps(content))


def serve(args) -> None:
    """
    Spin up a process listening for calls messages from the QGIS plugin.
    """
    HOST = "localhost"
    PORT = args.port[0]
    print(f"Starting TimServer on localhost, port: {PORT}")
    print(f"Process ID: {os.getpid()}")
    # Create the server, binding to localhost on port 9999
    with gistim.StatefulTimServer((HOST, PORT), gistim.TimHandler) as server:
        # Activate the server; this will keep running until you
        # interrupt the program with Ctrl-C
        server.serve_forever()


def extract(args) -> None:
    """
    Extract layer input from a netcdf dataset
    """
    inpath = args.inpath[0]
    outpath = args.outpath[0]
    wkt_geometry = args.wkt[0].split(";")
    gistim.data_extraction.netcdf_to_table(inpath, outpath, wkt_geometry)


def convert(args) -> None:
    """
    Convert a Geopackage into a Python script.
    """
    inpath = args.inpath[0]
    outpath = args.outpath[0]
    gistim.convert_to_script(inpath, outpath)


if __name__ == "__main__":
    # Setup argparsers
    parser = argparse.ArgumentParser(prog="GisTim")
    subparsers = parser.add_subparsers(help="sub-command help")
    parser_configure = subparsers.add_parser("configure", help="configure help")
    parser_serve = subparsers.add_parser("serve", help="serve help")
    parser_extract = subparsers.add_parser("extract", help="extract help")
    parser_convert = subparsers.add_parser("convert", help="convert help")

    parser_configure.set_defaults(func=configure)
    parser_configure.add_argument("append", type=bool, nargs=1, help="append")

    # Serve has a single argument: the port number
    parser_serve.set_defaults(func=serve)
    parser_serve.add_argument(
        "port",
        type=int,
        nargs=1,
        help="localhost PORT number",
    )

    parser_extract.set_defaults(func=extract)
    parser_extract.add_argument("inpath", type=str, nargs=1, help="inpath")
    parser_extract.add_argument("outpath", type=str, nargs=1, help="outpath")
    parser_extract.add_argument("wkt", type=str, nargs=1, help="wkt")

    parser_convert.set_defaults(func=convert)
    parser_convert.add_argument("inpath", type=str, nargs=1, help="inpath")
    parser_convert.add_argument("outpath", type=str, nargs=1, help="outpath")

    # Parse and call the appropriate function
    args = parser.parse_args()
    args.func(args)



from gistim import timml_elements, ttim_elements

from .common import gridspec, model_specification
from .data_extraction import as_aquifer_aquitard, layer_statistics
from .tim_server import StatefulTimServer, TimHandler
from .ugrid import to_ugrid2d


def convert_to_script(inpath: str, outpath: str) -> None:
    timml_spec, ttim_spec = model_specification(inpath, {})
    timml_script = timml_elements.convert_to_script(timml_spec)
    ttim_script = ttim_elements.convert_to_script(ttim_spec)

    with open(outpath, "w") as f:
        f.write(timml_script)
        f.write("\n")
        f.write(ttim_script)
